@@ -1,6 +1,6 @@
 #!/bin/bash
 # Copyright 2012-2019 Valve Software
-# Copyright 2019 Collabora Ltd.
+# Copyright 2019-2021 Collabora Ltd.
 #
 # SPDX-License-Identifier: MIT
 
@@ -8,6 +8,9 @@ set -e
 set -u
 set -o pipefail
 
+identify_library_abi=
+libcurl_compat_setup=
+
 # Check if set, which is normally done by steam.sh, but will not be set when invoked directly
 if [ -z "${STEAM_ZENITY+x}" ]; then
     # We are likely outside of run.sh as well, only use the host zenity
@@ -15,6 +18,26 @@ if [ -z "${STEAM_ZENITY+x}" ]; then
     export STEAM_ZENITY
 fi
 
+# If steam.sh or run.sh doesn't find a system zenity, it sets
+# STEAM_ZENITY=zenity, meaning do a PATH search and hope to find the one
+# from the Steam Runtime. However, in this script we don't expect the
+# Steam Runtime search paths to be available yet, so double-check: we
+# might find that no zenity is available.
+if ! [ -x "$(command -v "${STEAM_ZENITY}" || true)" ]; then
+    STEAM_ZENITY=
+fi
+
+log () {
+    echo "setup.sh[$$]: $*" >&2 || :
+}
+
+debug ()
+{
+    if [ -n "${STEAM_RUNTIME_VERBOSE-}" ]; then
+        log "$@"
+    fi
+}
+
 pin_newer_runtime_libs ()
 {
     # Set separator to newline just for this function
@@ -25,7 +48,7 @@ pin_newer_runtime_libs ()
     local zenity_progress=true
 
     # First argument is the runtime path
-    steam_runtime_path=$(realpath "$1")
+    steam_runtime_path=$(readlink -f "$1")
 
     if [[ ! -d "$steam_runtime_path" ]]; then return; fi
 
@@ -39,11 +62,11 @@ pin_newer_runtime_libs ()
     local -A host_libraries_64
 
     local bitness
-    local file_output
+    local soname_details
     local final_library
     local find_num
-    local find_output
-    local find_output_array
+    local line
+    local output_array
     local h_lib_major
     local h_lib_minor
     local h_lib_third
@@ -63,6 +86,8 @@ pin_newer_runtime_libs ()
     local soname_fullpath
     local soname_symlink
 
+    rm -rf "$steam_runtime_path/libcurl_compat_32"
+    rm -rf "$steam_runtime_path/libcurl_compat_64"
     rm -rf "$steam_runtime_path/pinned_libs_32"
     rm -rf "$steam_runtime_path/pinned_libs_64"
 
@@ -74,93 +99,171 @@ pin_newer_runtime_libs ()
     if [ "$zenity_progress" = true ]; then
         echo 4
     else
-        printf '\r 4%%    \r'
+        printf '\r 4%%    \r' >&2 || :
     fi
 
-    # First, grab the list of system libraries from ldconfig and put them in the arrays
-    shopt -s lastpipe
-    if /sbin/ldconfig -XNv 2> /dev/null | mapfile -t ldconfig_output_array; then
-        true  # OK
+    if [[ -n "$identify_library_abi" ]]; then
+        local executable_output_array
+        local line
+        local soname_path
+        local abi
+
+        # This usually takes only a few dozen milliseconds, there is no need
+        # to update the progress bar
+        mapfile -t executable_output_array < <($identify_library_abi --ldconfig --skip-unversioned 2>/dev/null)
+
+        for line in "${executable_output_array[@]}"
+        do
+            soname_path=${line%=*}
+            soname=${soname_path##*/}
+            abi=${line#*=}
+
+            if [[ ${abi} == "i386-linux-gnu" ]]; then
+                if [[ -n "${host_libraries_32[$soname]+isset}" ]]; then
+                    debug "Ignoring $soname_path because ${host_libraries_32[$soname]} is higher-precedence"
+                else
+                    host_libraries_32[$soname]=$soname_path
+                fi
+            elif [[ ${abi} == "x86_64-linux-gnu" ]]; then
+                if [[ -n "${host_libraries_64[$soname]+isset}" ]]; then
+                    debug "Ignoring $soname_path because ${host_libraries_64[$soname]} is higher-precedence"
+                else
+                    host_libraries_64[$soname]=$soname_path
+                fi
+            fi
+            # If it's not i386-linux-gnu nor x86_64-linux-gnu we just skip it
+        done
     else
-        >&2 echo "Warning: An unexpected error occurred while executing \"/sbin/ldconfig -XNv\", the exit status was $?"
+        # Fallback to the older, and slower, method that uses `file -L`
+
+        # First, grab the list of system libraries from ldconfig and put them in the arrays
+        shopt -s lastpipe
+        if /sbin/ldconfig -XNv 2> /dev/null | mapfile -t ldconfig_output_array; then
+            true  # OK
+        else
+            log "Warning: An unexpected error occurred while executing \"/sbin/ldconfig -XNv\", the exit status was $?"
+        fi
+
+        ldconfig_num=${#ldconfig_output_array[@]}
+        n_done=0
+
+        for ldconfig_output in "${ldconfig_output_array[@]}"
+        do
+            if [ "$zenity_progress" = true ]; then
+                echo $(( ( 60 * n_done ) / ldconfig_num + 4 ))
+            else
+                printf '\r %d%%    \r' $(( ( 60 * n_done ) / ldconfig_num + 4 )) >&2 || :
+            fi
+            (( n_done=n_done+1 ))
+            # If line starts with a leading / and contains :, it's a new path prefix
+            if [[ "$ldconfig_output" =~ ^/.*: ]]
+            then
+                library_path_prefix=$(echo "$ldconfig_output" | cut -d: -f1)
+            else
+                # Otherwise it's a soname symlink -> library pair, build a full path to the soname link
+                leftside=${ldconfig_output% -> *}
+                soname=$(echo "$leftside" | tr -d '[:space:]')
+                soname_fullpath=$library_path_prefix/$soname
+
+                # Left side better be a symlink
+                if [[ ! -L $soname_fullpath ]]; then continue; fi
+
+                # Left-hand side of soname symlink should be *.so.%d
+                if [[ ! $soname_fullpath =~ .*\.so.[[:digit:]]+$ ]]; then continue; fi
+
+                if ! final_library=$(readlink -f "$soname_fullpath")
+                then
+                    continue
+                fi
+
+                # Target library must be named *.so.%d.%d.%d
+                if [[ ! $final_library =~ .*\.so.[[:digit:]]+.[[:digit:]]+.[[:digit:]]+$ ]]; then continue; fi
+
+                # If it doesn't exist, skip as well
+                if [[ ! -f $final_library ]]; then continue; fi
+
+                soname_details=$(file -L "$final_library")
+
+                # Save into bitness-specific associative array with only SONAME as left-hand
+                if [[ $soname_details == *"32-bit"* ]]
+                then
+                    if [[ -n "${host_libraries_32[$soname]+isset}" ]]; then
+                        debug "Ignoring $soname_fullpath because ${host_libraries_32[$soname]} is higher-precedence"
+                    else
+                        host_libraries_32[$soname]=$soname_fullpath
+                    fi
+                elif [[ $soname_details == *"64-bit"* ]]
+                then
+                    if [[ -n "${host_libraries_64[$soname]+isset}" ]]; then
+                        debug "Ignoring $soname_fullpath because ${host_libraries_64[$soname]} is higher-precedence"
+                    else
+                        host_libraries_64[$soname]=$soname_fullpath
+                    fi
+                fi
+            fi
+        done
     fi
 
-    ldconfig_num=${#ldconfig_output_array[@]}
-    n_done=0
-
-    for ldconfig_output in "${ldconfig_output_array[@]}"
-    do
-        if [ "$zenity_progress" = true ]; then
-            echo $(( ( 60 * n_done ) / ldconfig_num + 4 ))
-        else
-            printf '\r %d%%    \r' $(( ( 60 * n_done ) / ldconfig_num + 4 ))
-        fi
-        (( n_done=n_done+1 ))
-        # If line starts with a leading / and contains :, it's a new path prefix
-        if [[ "$ldconfig_output" =~ ^/.*: ]]
-        then
-            library_path_prefix=$(echo "$ldconfig_output" | cut -d: -f1)
-        else
-            # Otherwise it's a soname symlink -> library pair, build a full path to the soname link
-            leftside=${ldconfig_output% -> *}
-            soname=$(echo "$leftside" | tr -d '[:space:]')
-            soname_fullpath=$library_path_prefix/$soname
-
-            # Left side better be a symlink
-            if [[ ! -L $soname_fullpath ]]; then continue; fi
-
-            # Left-hand side of soname symlink should be *.so.%d
-            if [[ ! $soname_fullpath =~ .*\.so.[[:digit:]]+$ ]]; then continue; fi
-
-            if ! final_library=$(readlink -f "$soname_fullpath")
-            then
-                continue
-            fi
-
-            # Target library must be named *.so.%d.%d.%d
-            if [[ ! $final_library =~ .*\.so.[[:digit:]]+.[[:digit:]]+.[[:digit:]]+$ ]]; then continue; fi
-
-            # If it doesn't exist, skip as well
-            if [[ ! -f $final_library ]]; then continue; fi
-
-            file_output=$(file -L "$final_library")
-
-            # Save into bitness-specific associative array with only SONAME as left-hand
-            if [[ $file_output == *"32-bit"* ]]
-            then
-                host_libraries_32[$soname]=$soname_fullpath
-            elif [[ $file_output == *"64-bit"* ]]
-            then
-                host_libraries_64[$soname]=$soname_fullpath
-            fi
-        fi
-    done
-
     if [ "$zenity_progress" = true ]; then
         echo 64
     else
-        printf '\r 64%%    \r'
+        printf '\r 64%%    \r' >&2 || :
     fi
 
     mkdir "$steam_runtime_path/pinned_libs_32"
     mkdir "$steam_runtime_path/pinned_libs_64"
 
-    mapfile -t find_output_array < <(find "$steam_runtime_path" -type l | grep \\\.so)
-    find_num=${#find_output_array[@]}
+    # Give steamrt a chance to fix libcurl ABI conflicts in a cleverer way.
+    # This will only work for glibc >= 2.30, but if it does work, it will
+    # create a libcurl_compat_${bitness}/libcurl.so.4 that is better than
+    # anything we can do from this shell script. This relies on tricky
+    # implementation details of glibc, so we're not using it by default yet
+    # (see run.sh for the opt-in mechanism).
+    if [[ -x "$libcurl_compat_setup" ]] && "$libcurl_compat_setup" --runtime-optional "$steam_runtime_path"; then
+        debug "run.sh can optionally use shim library to support more than one libcurl ABI"
+    fi
+
+    if [[ -n "$identify_library_abi" ]]; then
+        mapfile -t output_array < <($identify_library_abi --directory "$steam_runtime_path" --skip-unversioned  2>/dev/null)
+    else
+        mapfile -t output_array < <(find "$steam_runtime_path" -type l | grep \\\.so)
+    fi
+
+    find_num=${#output_array[@]}
     n_done=0
 
-    for find_output in "${find_output_array[@]}"
+    for line in "${output_array[@]}"
     do
+        case "$line" in
+            (*/*-linux-gnu/sse2/lib* | */*-linux-gnu/sse/lib* | */*-linux-gnu/i686/cmov/lib*)
+                debug "Library in hwcap directory: $line"
+                ;;
+            (*/*-linux-gnu/lib*)
+                debug "Library in base directory: $line"
+                ;;
+            (*)
+                debug "Skip library in subdirectory: $line"
+                continue
+                ;;
+        esac
+
         if [ "$zenity_progress" = true ]; then
             echo $(( ( 35 * n_done ) / find_num + 64 ))
         else
-            printf '\r %d%%    \r' $(( ( 35 * n_done ) / find_num + 64 ))
+            printf '\r %d%%    \r' $(( ( 35 * n_done ) / find_num + 64 )) >&2 || :
         fi
         (( n_done=n_done+1 ))
-        # Left-hand side of soname symlink should be *.so.%d
-        if [[ ! $find_output =~ .*\.so.[[:digit:]]+$ ]]; then continue; fi
 
-        soname_symlink=$find_output
+        if [[ -n "$identify_library_abi" ]]; then
+            soname_symlink=${line%=*}
+            soname_details=${line#*=}
+        else
+            soname_symlink=$line
+            soname_details=
+        fi
+
+        # Left-hand side of soname symlink should be *.so.%d
+        if [[ ! $soname_symlink =~ .*\.so.[[:digit:]]+$ ]]; then continue; fi
 
         if ! final_library=$(readlink -f "$soname_symlink")
         then
@@ -182,60 +285,32 @@ pin_newer_runtime_libs ()
         host_soname_symlink=""
         bitness="unknown"
 
-        soname=$(basename "$soname_symlink")
+        soname=${soname_symlink##*/}
 
-        # If we had entries in our arrays, get them
-        if [[ -n "${host_libraries_32[$soname]+isset}" ||
-              -n "${host_libraries_64[$soname]+isset}" ]]
+        if [[ -z $soname_details ]]; then
+            soname_details=$(file -L "$final_library")
+        fi
+
+        if [[ $soname_details == *"32-bit"* || $soname_details == "i386-linux-gnu" ]]
         then
-            file_output=$(file -L "$final_library")
-            if [[ $file_output == *"32-bit"* ]]
+            if [ -n "${host_libraries_32[$soname]+isset}" ]
             then
-                if [ -n "${host_libraries_32[$soname]+isset}" ]
-                then
-                    host_soname_symlink=${host_libraries_32[$soname]}
-                fi
-                bitness="32"
-            elif [[ $file_output == *"64-bit"* ]]
-            then
-                if [ -n "${host_libraries_64[$soname]+isset}" ]
-                then
-                    host_soname_symlink=${host_libraries_64[$soname]}
-                fi
-                bitness="64"
+                host_soname_symlink=${host_libraries_32[$soname]}
             fi
-       fi
+            bitness="32"
+        elif [[ $soname_details == *"64-bit"* || $soname_details == "x86_64-linux-gnu" ]]
+        then
+            if [ -n "${host_libraries_64[$soname]+isset}" ]
+            then
+                host_soname_symlink=${host_libraries_64[$soname]}
+            fi
+            bitness="64"
+        fi
 
-        # Do we have a host library found for the same SONAME?
-        if [[ ! -f $host_soname_symlink || $bitness == "unknown" ]]; then continue; fi
-
-        host_library=$(readlink -f "$host_soname_symlink")
-
-        if [[ ! -f $host_library ]]; then continue; fi
-
-        #echo $soname ${host_libraries[$soname]} $r_lib_major $r_lib_minor $r_lib_third
-
-        # Pretty sure the host library already matches, but we need the rematch anyway
-        if [[ ! $host_library =~ .*\.so.([[:digit:]]+).([[:digit:]]+).([[:digit:]]+)$ ]]; then continue; fi
-
-        h_lib_major=$((10#${BASH_REMATCH[1]}))
-        h_lib_minor=$((10#${BASH_REMATCH[2]}))
-        h_lib_third=$((10#${BASH_REMATCH[3]}))
+        if [[ $bitness == "unknown" ]]; then continue; fi
 
         runtime_version_newer="no"
 
-        if [[ $h_lib_major -lt $r_lib_major ]]; then
-            runtime_version_newer="yes"
-        fi
-
-        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -lt $r_lib_minor ]]; then
-            runtime_version_newer="yes"
-        fi
-
-        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -eq $r_lib_minor && $h_lib_third -lt $r_lib_third ]]; then
-            runtime_version_newer="yes"
-        fi
-
         # There's a set of libraries that have to work together to yield a working dock
         # We're reasonably convinced our set works well, and only pinning a handful would
         # induce a mismatch and break the dock, so always pin all of these for Steam (32-bit)
@@ -270,20 +345,76 @@ pin_newer_runtime_libs ()
                 # built against the Steam Runtime will expect the old SONAME
                 # and versioned symbols; make sure we use the Steam Runtime
                 # version.
+                #
+                # (If steamrt-libcurl-compat already fixed this for us, then
+                # it will have created the libcurl.so.4 symlink and as a
+                # result we'll never get here.)
+                runtime_version_newer="forced"
+                ;;
+
+            (libcurl-gnutls.so.4)
+                # Similar to the above, the GNUTLS variant of libcurl has
+                # Debian-specific symbol-versioning. We didn't have a
+                # problem with this until recently, because most OSs either
+                # don't ship a GNUTLS variant at all (Red Hat), ship it
+                # without versioned symbols (Arch), or are Debian-derived
+                # (Debian/Ubuntu), but apparently OpenMandriva has a
+                # non-Debian-compatible GNUTLS variant.
                 runtime_version_newer="forced"
                 ;;
         esac
 
-        # Print to stderr because zenity is consuming stdout
-        if [[ $runtime_version_newer == "yes" ]]; then
-            >&2 echo "Found newer runtime version for $bitness-bit $soname. Host: $h_lib_major.$h_lib_minor.$h_lib_third Runtime: $r_lib_major.$r_lib_minor.$r_lib_third"
-        elif [[ $runtime_version_newer == "forced" ]]; then
-            >&2 echo "Forced use of runtime version for $bitness-bit $soname. Host: $h_lib_major.$h_lib_minor.$h_lib_third Runtime: $r_lib_major.$r_lib_minor.$r_lib_third"
+        # If the library is one of the ones we force, we need to
+        # do that even if we are not aware of an equivalent on the host
+        # system, because continuing uncertainty about the ABIs of these
+        # libraries means that it might have an unexpected SONAME.
+        # For instance, some versions of OpenMandriva's lib64curl-gnutls4
+        # contained a libcurl-gnutls.so.4 with SONAME libcurl.so.4.
+        if [ "$runtime_version_newer" = forced ]; then
+            log "Forced use of runtime version for $bitness-bit $soname"
+            ln -fns "$final_library" "$steam_runtime_path/pinned_libs_$bitness/$soname"
+            # For similar historical reasons, our libcurl*.so.3 are equivalent
+            # to libcurl*.so.4
+            case "$soname" in
+                (libcurl*.so.4)
+                    ln -fns "$final_library" "$steam_runtime_path/pinned_libs_$bitness/${soname%.4}.3"
+                    ;;
+            esac
+            continue
         fi
 
-        if [[ $runtime_version_newer == "yes" \
-              || $runtime_version_newer == "forced" ]]; then
-            ln -s "$final_library" "$steam_runtime_path/pinned_libs_$bitness/$soname"
+        # Do we have a host library found for the same SONAME?
+        if [[ ! -f $host_soname_symlink ]]; then continue; fi
+
+        host_library=$(readlink -f "$host_soname_symlink")
+
+        if [[ ! -f $host_library ]]; then continue; fi
+
+        #log $soname ${host_libraries[$soname]} $r_lib_major $r_lib_minor $r_lib_third
+
+        # Pretty sure the host library already matches, but we need the rematch anyway
+        if [[ ! $host_library =~ .*\.so.([[:digit:]]+).([[:digit:]]+).([[:digit:]]+)$ ]]; then continue; fi
+
+        h_lib_major=$((10#${BASH_REMATCH[1]}))
+        h_lib_minor=$((10#${BASH_REMATCH[2]}))
+        h_lib_third=$((10#${BASH_REMATCH[3]}))
+
+        if [[ $h_lib_major -lt $r_lib_major ]]; then
+            runtime_version_newer="yes"
+        fi
+
+        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -lt $r_lib_minor ]]; then
+            runtime_version_newer="yes"
+        fi
+
+        if [[ $h_lib_major -eq $r_lib_major && $h_lib_minor -eq $r_lib_minor && $h_lib_third -lt $r_lib_third ]]; then
+            runtime_version_newer="yes"
+        fi
+
+        # Print to stderr because zenity is consuming stdout
+        if [[ $runtime_version_newer == "yes" ]]; then
+            log "Found newer runtime version for $bitness-bit $soname. Host: $h_lib_major.$h_lib_minor.$h_lib_third Runtime: $r_lib_major.$r_lib_minor.$r_lib_third"
+            ln -fns "$final_library" "$steam_runtime_path/pinned_libs_$bitness/$soname"
             # Keep track of the exact version name we saw on the system at pinning time to check later
             >&2 echo "$host_soname_symlink" > "$steam_runtime_path/pinned_libs_$bitness/system_$soname"
             >&2 echo "$host_library" >> "$steam_runtime_path/pinned_libs_$bitness/system_$soname"
@@ -292,24 +423,13 @@ pin_newer_runtime_libs ()
     done
 
     for bitness in 32 64; do
-        if [ -L "$steam_runtime_path/pinned_libs_$bitness/libcurl.so.4" ]; then
-            # The version of libcurl.so.4 in the Steam Runtime is actually
-            # binary-compatible with the older libcurl.so.3 in Debian/Ubuntu,
-            # so pin it under both names.
-            ln -fns libcurl.so.4 "$steam_runtime_path/pinned_libs_$bitness/libcurl.so.3"
-        fi
-
-        if [ -L "$steam_runtime_path/pinned_libs_$bitness/libcurl-gnutls.so.4" ]; then
-            # Similarly, libcurl-gnutls.so.4 is actually binary-compatible
-            # with libcurl-gnutls.so.3, so pin it under both names.
-            ln -fns libcurl-gnutls.so.4 "$steam_runtime_path/pinned_libs_$bitness/libcurl-gnutls.so.3"
-        fi
+        touch "$steam_runtime_path/pinned_libs_$bitness/done"
     done
 
     if [ "$zenity_progress" = true ]; then
         echo 100
     else
-        echo " 100%   "
+        echo " 100%   " >&2 || :
     fi
 }
 
@@ -326,14 +446,15 @@ check_pins ()
     local steam_runtime_path
 
     # First argument is the runtime path
-    steam_runtime_path=$(realpath "$1")
+    steam_runtime_path=$(readlink -f "$1")
 
     if [[ ! -d "$steam_runtime_path" ]]; then return; fi
 
     pins_need_redoing="no"
 
-    # If we had the runtime previously unpacked but never ran the pin code, do it now
-    if [[ ! -d "$steam_runtime_path/pinned_libs_32" ]]
+    # If we had the runtime previously unpacked but never ran the pin code,
+    # or if a previous attempt failed or was cancelled, do it now
+    if ! [[ -e "$steam_runtime_path/pinned_libs_32/done" && -e "$steam_runtime_path/pinned_libs_64/done" ]]
     then
         pins_need_redoing="yes"
     fi
@@ -371,7 +492,7 @@ check_pins ()
 
     if [[ $pins_need_redoing == "yes" ]]
     then
-        echo Updating Steam runtime environment...
+        log "Updating Steam runtime environment..."
         # Is always set at this point, but may be empty if the host lacks zenity
         if [ -n "${STEAM_ZENITY}" ]; then
             pin_newer_runtime_libs "$steam_runtime_path" | "${STEAM_ZENITY}" --progress --auto-close --percentage=0 --no-cancel --width 400 --title="Steam setup" --text="Updating Steam runtime environment..."
@@ -379,7 +500,7 @@ check_pins ()
             pin_newer_runtime_libs "$steam_runtime_path" "false"
         fi
     else
-        echo Steam runtime environment up-to-date!
+        log "Steam runtime environment up-to-date!"
     fi
 }
 
@@ -458,7 +579,7 @@ main ()
                 ;;
 
             (-*)
-                echo "$me: unknown option: $1" >&2
+                log "unknown option: $1"
                 usage 2
                 ;;
 
@@ -470,21 +591,32 @@ main ()
 
     if [ "$#" -gt 0 ]
     then
-        echo "$me does not have positional parameters" >&2
+        log "$me does not have positional parameters"
         usage 2
     fi
 
+    case "$(uname -m)" in
+        (*64)
+            arch=amd64
+            ;;
+        (*)
+            arch=i386
+            ;;
+    esac
+
+    identify_library_abi="$top/$arch/usr/bin/steam-runtime-identify-library-abi"
+    libcurl_compat_setup="$top/$arch/usr/bin/steam-runtime-libcurl-compat-setup"
+
+    if [[ ! -x "$identify_library_abi" ]]; then
+        identify_library_abi="$top/$arch/bin/steam-runtime-identify-library-abi"
+        if [[ ! -x "$identify_library_abi" ]]; then
+            identify_library_abi=
+        fi
+    fi
+
     case "$action" in
         (--print-bin-path)
             # Keep this in sync with run.sh
-            case "$(uname -m)" in
-                (*64)
-                    arch=amd64
-                    ;;
-                (*)
-                    arch=i386
-                    ;;
-            esac
             echo "$top/$arch/bin:$top/$arch/usr/bin:$top/usr/bin"
             ;;
 
